



library(foreign)



doIt <- function(d,marker,riskh=NULL,riskl=NULL) {
  n = length(d) 

  mod       <- glm(d~marker,family=binomial("logit"))
  fitVal    <- mod$fitted.values
  perc      <- rank(fitVal,ties="max")/n

  reorder   <- order(fitVal)
  fitVal    <- fitVal[reorder]
  perc      <- perc[reorder]
  d         <- d[reorder]
  whichCase <- which(d==1)
  numCase   <- length(whichCase)
  whichCont <- which(d==0)
  numCont   <- length(whichCont)

  rankAll   <- rank(fitVal,ties="min")
  rankCase  <- rank(fitVal[whichCase],ties="min")
  rankCont  <- rank(fitVal[whichCont],ties="min")

  tprAll            = rep(NA,n)
  fprAll            = rep(NA,n)
  tprAll[whichCase] = (1 + numCase - rankCase)/numCase
  tprAll[whichCont] = (numCase - rankAll[whichCont] + rankCont)/numCase
  fprAll[whichCont] = (1 + numCont - rankCont)/numCont
  fprAll[whichCase] = (numCont - rankAll[whichCase] + rankCase)/numCont

  riskHperc = NULL
  riskLperc = NULL
  tprH      = NULL
  tprL      = NULL
  fprH      = NULL
  fprL      = NULL

  if(!is.null(riskh)) {
    riskHPos  <- min(which(fitVal > riskh))
    riskHperc <- 1 - perc[riskHPos]
    tprH      <- tprAll[riskHPos]
    fprH      <- fprAll[riskHPos]
  }

  if(!is.null(riskl)) {
    riskLPos  <- max(which(fitVal <= riskl))
    riskLperc <- perc[riskLPos]
    tprL      <- tprAll[riskLPos]
    fprL      <- fprAll[riskLPos]
  }
  
  list(fitted = fitVal, perc = perc, fpr = fprAll, tpr = tprAll,
       whichCase=whichCase,whichCont=whichCont, 
       riskHperc=riskHperc,riskLperc=riskLperc,
       tprH=tprH,tprL=tprL, fprH=fprH,fprL=fprL)
}

plotFPRTPR <- function(listRet,new=TRUE) {
  if(new) {
    x= c(0,1)
    y= x
    plot(y~x,type="n")
  }
  lines(listRet$tpr~listRet$perc)
  lines(listRet$fpr~listRet$perc)  
}


mydata <- read.table("http://labs.fhcrc.org/pepe/data/janssens_c.csv",sep=",",header=T)

m1 <- doIt(mydata$d,mydata$logscr,riskh=.4,riskl=.1)
m2 <- doIt(mydata$d,mydata$bmi,riskh=.4,riskl=.1)


par(mfrow=c(2,1))
plot(m1$fitted~m1$perc,type="l",col="green")
lines(m2$fitted~m2$perc,type="l",col="orange")

plotFPRTPR(m1)
plotFPRTPR(m2,new=FALSE)


tpf1 <- 

setwd("/Users/darylm/Desktop/pcvsuite")
source("R/hidden.R")

janssens <- read.table("janssens_c.csv",sep=",",header=T)



dataset="janssens"; d="disease"; markers=c("logscr")
covar=NULL; link="logit"; riskl=NULL; riskh=NULL; ci=FALSE; rho=NULL 
   ylim=NULL; class=FALSE; offset=.005; class_offset=.005
   nsamp=1000; cluster=NULL; level=95; bootseed=51

markers=c("logscr","bmi");  riskh=.4; riskl = .1;  class=TRUE  
   

dataset="mydata";  d="disease"; markers=c("logscr","bmi"); covar=NULL; link="logit"
   riskl=.1; riskh=.4; ci=TRUE; rho=NULL 
   ylim=NULL; class=TRUE; offset=.01; class_offset=.01
   nsamp=1000; cluster=NULL; level=95; genprvvars=FALSE
   
markers=c("logscr","bmi"); 
covar=c("age","hypertension","bmi","bruit","vascular","gender")

source("R/predcurve.R") 

predcurve(dataset="mydata",d="disease", markers="logscr", 
   covar=c("age","hypertension","bmi","bruit","vascular","gender"), 
   link="logit", riskl=.1, riskh=.4, ci=TRUE, class=TRUE)
   